;
; Stack and Heap Initialization
;
; Implement the function __user_setup_stackheap() for AArch64
;
; Supports 2 clusters, each with up to 2 cores
;

    PRESERVE8
    AREA  BOOT, CODE, READONLY, ALIGN=3

    EXPORT __user_setup_stackheap

    IMPORT ||Image$$HEAP00$$ZI$$Base||
    IMPORT ||Image$$HEAP00$$ZI$$Limit||
    IMPORT ||Image$$HEAP01$$ZI$$Base||
    IMPORT ||Image$$HEAP01$$ZI$$Limit||
    IMPORT ||Image$$HEAP10$$ZI$$Base||
    IMPORT ||Image$$HEAP10$$ZI$$Limit||
    IMPORT ||Image$$HEAP11$$ZI$$Base||
    IMPORT ||Image$$HEAP11$$ZI$$Limit||

    ; We override __user_setup_stackheap so that the C library won't set stack/heap

__user_setup_stackheap
    ; - heap base goes in X0
    ; - heap limit goes in X2

    MRS x0, MPIDR_EL1 
    AND x0, x0, #0x1FF
    CMP x0, #0x0
    BEQ CLUS0CPU0
    CMP x0, #0x1
    BEQ CLUS0CPU1
    CMP x0, #0x100
    BEQ CLUS1CPU0
    CMP x0, #0x101
    BEQ CLUS1CPU1
    B   DONE

CLUS0CPU0
    LDR     X0, =||Image$$HEAP00$$ZI$$Base||
    LDR     X2, =||Image$$HEAP00$$ZI$$Limit||
    B       DONE

CLUS0CPU1
    LDR     X0, =||Image$$HEAP01$$ZI$$Base||
    LDR     X2, =||Image$$HEAP01$$ZI$$Limit||
    B       DONE

CLUS1CPU0
    LDR     X0, =||Image$$HEAP10$$ZI$$Base||
    LDR     X2, =||Image$$HEAP10$$ZI$$Limit||
    B       DONE

CLUS1CPU1
    LDR     X0, =||Image$$HEAP11$$ZI$$Base||
    LDR     X2, =||Image$$HEAP11$$ZI$$Limit||

DONE
    RET

    END

