;==================================================================
; Copyright ARM Ltd 2012. All rights reserved.
;
; ARMv8 example - Startup Code
;==================================================================

        PRESERVE8
        AREA  BOOT, CODE, READONLY, ALIGN=3 

        ENTRY


TT_S1_TABLE              EQU    0x00000000000000003    ; NSTable=0, PXNTable=0, UXNTable=0, APTable=0

; TT block entries templates   (L1 and L2, NOT L3)    
; Assuming table contents:
; 0 = b01000100 = Normal, Inner/Outer Non-Cacheable
; 1 = b11111111 = Normal, Inner/Outer WB/WA/RA
; 2 = b00000000 = Device-nGnRnE
TT_S1_FAULT              EQU    0x0
TT_S1_NORMAL_NO_CACHE    EQU    0x00000000000000401    ; Index = 0, AF=1
TT_S1_NORMAL_WBWA        EQU    0x00000000000000405    ; Index = 1, AF=1
TT_S1_DEVICE_nGnRnE      EQU    0x00000000000000409    ; Index = 2, AF=1

TT_S1_UXN                EQU    (1 << 54)
TT_S1_PXN                EQU    (1 << 53)
TT_S1_nG                 EQU    (1 << 11)
TT_S1_NS                 EQU    (1 << 5)

TT_S1_NON_SHARED         EQU    (0 << 8)               ; Non-shareable
TT_S1_INNER_SHARED       EQU    (3 << 8)               ; Inner-shareable
TT_S1_OUTER_SHARED       EQU    (2 << 8)               ; Outer-shareable

TT_S1_PRIV_RW            EQU    (0x0)
TT_S1_PRIV_RO            EQU    (0x2 << 6)
TT_S1_USER_RW            EQU    (0x1 << 6)
TT_S1_USER_RO            EQU    (0x3 << 6)

TT_FLASH                 EQU    TT_S1_NORMAL_WBWA :OR: TT_S1_INNER_SHARED :OR: TT_S1_PRIV_RO
TT_RAM                   EQU    TT_S1_NORMAL_WBWA :OR: TT_S1_INNER_SHARED :OR: TT_S1_PRIV_RW
TT_PERIPHERAL            EQU    TT_S1_DEVICE_nGnRnE :OR: TT_S1_PXN :OR: TT_S1_UXN :OR: TT_S1_PRIV_RW
;==================================================================
; Entry point 
;==================================================================


    EXPORT StartHere
StartHere

    MOV X0, XZR
    MOV X1, XZR
    MOV X2, XZR
    MOV X3, XZR
    MOV X4, XZR
    MOV X5, XZR
    MOV X6, XZR
    MOV X7, XZR
    MOV X8, XZR
    MOV X9, XZR
    MOV X10, XZR
    MOV X11, XZR
    MOV X12, XZR
    MOV X13, XZR
    MOV X14, XZR
    MOV X15, XZR
    MOV X16, XZR
    MOV X17, XZR
    MOV X18, XZR
    MOV X19, XZR
    MOV X20, XZR
    MOV X21, XZR
    MOV X22, XZR
    MOV X23, XZR
    MOV X24, XZR
    MOV X25, XZR
    MOV X26, XZR
    MOV X27, XZR
    MOV X28, XZR
    MOV X29, XZR
    MOV X30, XZR

;==================================================================
; Disable caches, MMU and branch prediction in case they were left enabled from an earlier run
; This does not need to be done from a cold reset
;==================================================================
    MRS x0, SCTLR_EL3            ; System Control Register (EL3)
    ORR x1, #(0x1 << 12)         ; Disable instruction cache
    ORR x1, #(0x1 << 2)          ; Disable data cache
    ORR x1, #(0x1 << 0)          ; Disable MMU
    BIC x0, x0, x1
    MSR SCTLR_EL3, x0
    ISB

; Caches are enabled inside main(),
; after scatterloading has been performed.

;==================================================================
; Initialize Monitor Mode Stack
; Note stack must be 128 byte aligned.
;==================================================================
    ;; Get the CPU ID bits [1:0] and Cluster ID [11:8] from MPIDR register and put it in r0
    ;; this initialization is currently setup for 1 or 2 clusters with 1 or 2 cpus 

    MRS x0,MPIDR_EL1 ;
    LSR x0, x0, #8
    AND x0, x0, #0xF
    CMP x0, #0x000
    BEQ STACKCPU00
    CMP x0, #0x001
    BEQ STACKCPU01
    CMP x0, #0x100
    BEQ STACKCPU10
    CMP x0, x1
    BEQ STACKCPU11

STACKCPU00
    IMPORT  ||Image$$STACK00$$ZI$$Limit|| ; Linker symbol from scatter file
    LDR     x0, =||Image$$STACK00$$ZI$$Limit||
    ADD     SP, x0, #0
    B      SPDONE
STACKCPU01
    IMPORT  ||Image$$STACK01$$ZI$$Limit|| ; Linker symbol from scatter file
    LDR     x0, =||Image$$STACK01$$ZI$$Limit||
    ADD     SP, x0, #0
    B      SPDONE
STACKCPU10
    IMPORT  ||Image$$STACK10$$ZI$$Limit|| ; Linker symbol from scatter file
    LDR     x0, =||Image$$STACK10$$ZI$$Limit||
    ADD     SP, x0, #0
    B      SPDONE
STACKCPU11
    IMPORT  ||Image$$STACK11$$ZI$$Limit|| ; Linker symbol from scatter file
    LDR     x0, =||Image$$STACK11$$ZI$$Limit||
    ADD     SP, x0, #0
SPDONE

;============================================================

; Set Vector Base Address Register (VBAR) to point to this application's vector table
;===================================================================
    LDR x0, =monitor_vectors     ; x0 = 0x00000800
    MSR VBAR_EL3, x0             ; EL3 sets vector base address
 
;==================================================================
; Enable VFP access from EL3
;==================================================================
    MRS x0, CPTR_EL3
    MOV w1, #0x400
    BIC w0, w0, w1
    MSR CPTR_EL3, x2

;==================================================================
; Only CPU0 continues to execute and the secondary CPUs execute a WFI instruction
;==================================================================
    MRS x0, MPIDR_EL1             ; Read Multiprocessor Affinity Register
    ANDS x0, x0, #0x03            ; Mask off, leaving the CPU ID field
    BNE  secondaryCPUsInit  

    ; Configure ACTLR_EL2
    ; --------------------
    ; These bits are IMP DEF, so need to different for different
    ; processors
    MRS      x1, ACTLR_EL2
    ORR      x1, x1, #1          ; Enable EL1 access to ACTLR_EL1
    ORR      x1, x1, #(1 << 1)   ; Enable EL1 access to CPUECTLR_EL1
    ORR      x1, x1, #(1 << 4)   ; Enable EL1 access to L2CTLR_EL1
    ORR      x1, x1, #(1 << 5)   ; Enable EL1 access to L2ECTLR_EL1
    ORR      x1, x1, #(1 << 6)   ; Enable EL1 access to L2ACTLR_EL1
    MSR      ACTLR_EL2, x1

    ; Configure CPUECTLR_EL1
    ; -----------------------
    ; These bits are IMP DEF, so need to different for different
    ; processors
    ; SMPEN - bit 6 - Enables the processor to receive cache
    ;                 and TLB maintenance operations
    ;
    ; NOTE: For Cortex-A57/53 CPUEN should be set beforebefore
    ;       enabling the caches and MMU, or performing any cache
    ;       and TLB maintenance operations.
    MRS      x0, S3_1_c15_c2_1  ; Read EL1 CPU Extended Control Register
    ORR      x0, x0, #(1 << 6)  ; Set the SMPEN bit
    MSR      S3_1_c15_c2_1, x0  ; Write EL1 CPU Extended Control Register
    ISB


; Set translation table base
;===================================================================

    MRS     x4, MPIDR_EL1
    LDR     x1, =0x101
    AND     x4, x4, x1


; Set the Base address
; ---------------------
    IMPORT   ||Image$$TT0_L1$$ZI$$Base||
    IMPORT   TT0_L2    
    LDR      x1, =||Image$$TT0_L1$$ZI$$Base||     ; Get address of level 1 for TTBR0_EL3
    MSR      TTBR0_EL3, x1                    ; Set TTBR0_EL3 (NOTE: There is no TTBR1 at EL3)

; Set up memory attributes
; -------------------------
; This equates to:
; 0 = b01000100 = Normal, Inner/Outer Non-Cacheable
; 1 = b11111111 = Normal, Inner/Outer WB/WA/RA
; 2 = b00000000 = Device-nGnRnE
    LDR      x1, =0x000000000000FF44
    MSR      MAIR_EL3, x1

; Set up TCR_EL3
; ---------------
; EPD0=0b0      Table walks using TTBR0 enabled
; TBI0=0b0      Top byte not ignored
; TG0=0b00      4KB granule
; T0SZ=0b011001 Limits VA space to 38 bits, translation starts @ l1
; SH0=0b11      Inner Shareable
; IGRN0=0b01    Walks to TTBR0 are Inner WB/WA
; OGRN0=0b01    Walks to TTBR0 are Outer WB/WA
    LDR      x1, =0x0000000080803519
    MSR      TCR_EL3, x1
  

; Ensure changes to system register are visible before MMU enabled
  ISB

;  L1 entries set to TT_S1_FAULT by default

; Now write in the entries we are going to use:
; table[0] = 0x0000,0000,0000,0000 - 0x0000,0000,3FFF,FFFF   (Peripherals, including UART0)
; table[1] = 0x0000,0000,4000,0000 - 0x0000,0000,7FFF,FFFF
; table[2] = 0x0000,0000,8000,0000 - 0x0000,0000,BFFF,FFFF   (RAM on the VE board model)
; table[3] = 0x0000,0000,C000,0000 - 0x0000,0000,FFFF,FFFF   (RAM on the VE board model)

    LDR      x0, =||Image$$TT0_L1$$ZI$$Base||  ; Address of L1 table
    LDR      x1, =TT0_L2
        
    LDR      x2, =TT_S1_TABLE                  ; Entry template
    ORR      x2, x2, x1                        ; Combine template with L2 table Base address
    STR      x2, [x0, #0]                      ; Write template into entry table[0]
    ADD      x2, x2, #0x1000                   ; Move pointer on to the next table in memory
    STR      x2, [x0, #8]                      ; Write template into entry table[1]
    ADD      x2, x2, #0x1000
    STR      x2, [x0, #16]                     ; Write template into entry table[2]
    ADD      x2, x2, #0x1000
    STR      x2, [x0, #24]                     ; Write template into entry table[3]
        
    DSB      SY


;===================================================================
; Branch to __main
; Leaving the caches disabled until after scatter loading.
;===================================================================

    MRS x1, SCTLR_EL3
    MOV     x2, #0x1805               
    ORR     x1, x1, x2
    MSR SCTLR_EL3, x1
    ISB

    IMPORT __main
    LDR      x0, =__main
    BR       x0
        
    B __main

;==================================================================
; Enable caches
;==================================================================
    AREA   ENABLECACHES, CODE, READONLY
    EXPORT enable_caches
enable_caches  FUNCTION
    IC IALLU                     ; Invalidate I cache and BTAC
    DSB SY
    ISB
    MRS x1, SCTLR_EL3
    MOV w2, #1
    BFI w1, w2, #12, #1          ; Enable instruction cache
    BFI w1, w2, #2, #1           ; Enable data cache
    MSR SCTLR_EL3, x1
    ISB

    RET

    ENDFUNC

;==================================================================
; Return CNTPCT_EL0 value
;==================================================================
    AREA   COUNTER, CODE, READONLY
    EXPORT get_count
get_count FUNCTION
    MRS x0, CNTPCT_EL0           ; Read CNTPCT_EL0 into X0
    RET
    ENDFUNC

;==================================================================
; Secondary CPU area
;==================================================================
    AREA   SECONDARY, CODE, READONLY
    EXPORT secondaryCPUsInit
secondaryCPUsInit FUNCTION
    WFI
    ENDFUNC

;==================================================================
; EL3 VECTOR TABLE
;==================================================================
    AREA  VECTORS, CODE, ALIGN = 12 
    EXPORT monitor_vectors
monitor_vectors
    B .      ; Current EL 32bits: Synchronous
    ALIGN 128
    B .      ;                    IRQ/vIRQ
    ALIGN 128
    B .      ;                    FIQ/vFIQ
    ALIGN 128
    B .      ;                    Error/vError
    ALIGN 128
    B .      ; Current EL 64bits: Synchronous
    ALIGN 128
    B .      ;                    IRQ/vIRQ
    ALIGN 128
    B .      ;                    FIQ/vFIQ
    ALIGN 128
    B .      ;                    Error/vError
    ALIGN 128
    B .      ; Lower EL SPx:      Synchronous
    ALIGN 128
    B .      ;                    IRQ/vIRQ
    ALIGN 128
    B .      ;                    FIQ/vFIQ
    ALIGN 128
    B .      ;                    Error/vError
    ALIGN 128
    B .      ; Lower EL SP0:      Synchronous
    ALIGN 128
    B .      ;                    IRQ/vIRQ
    ALIGN 128
    B .      ;                    FIQ/vFIQ
    ALIGN 128
    B .      ;                    Error/vError

    END


